use std::{
    cell::UnsafeCell,
    marker::PhantomPinned,
    mem::{self, ManuallyDrop},
    panic::{AssertUnwindSafe, catch_unwind},
    pin::Pin,
    ptr::NonNull,
    sync::{
        Arc, OnceLock,
        atomic::{AtomicBool, Ordering},
    },
};

use alloc::collections::BTreeMap;

use async_task::Runnable;

use crate::{
    channel::{Parker, Sender},
    heartbeat::HeartbeatList,
    job::{HeapJob, Job2 as Job, SharedJob, StackJob},
    queue::ReceiverToken,
    util::DropGuard,
    workerthread::{HeartbeatThread, WorkerThread},
};

pub struct Context {
    should_exit: AtomicBool,
    pub heartbeats: HeartbeatList,
    pub(crate) queue: Arc<crate::queue::Queue<Message>>,
    pub(crate) heartbeat: Parker,
}

pub(crate) enum Message {
    Shared(SharedJob),
    WakeUp,
    Exit,
    ScopeFinished,
}

pub(crate) struct Shared {
    pub jobs: BTreeMap<usize, SharedJob>,
    injected_jobs: Vec<SharedJob>,
}

unsafe impl Send for Shared {}

impl Shared {
    pub fn pop_job(&mut self) -> Option<SharedJob> {
        // this is unlikely, so make the function cold?
        // TODO: profile this
        if !self.injected_jobs.is_empty() {
            // SAFETY: we checked that injected_jobs is not empty
            unsafe { return Some(self.pop_injected_job()) };
        } else {
            self.jobs.pop_first().map(|(_, job)| job)
        }
    }

    #[cold]
    unsafe fn pop_injected_job(&mut self) -> SharedJob {
        self.injected_jobs.pop().unwrap()
    }
}

impl Context {
    pub fn new_with_threads(num_threads: usize) -> Arc<Self> {
        #[cfg(feature = "tracing")]
        tracing::trace!("Creating context with {} threads", num_threads);

        let this = Arc::new(Self {
            should_exit: AtomicBool::new(false),
            heartbeats: HeartbeatList::new(),
            queue: crate::queue::Queue::new(),
            heartbeat: Parker::new(),
        });

        // Create a barrier to synchronize the worker threads and the heartbeat thread
        let barrier = Arc::new(std::sync::Barrier::new(num_threads + 2));

        for i in 0..num_threads {
            let ctx = this.clone();
            let barrier = barrier.clone();

            std::thread::Builder::new()
                .name(format!("worker-{}", i))
                .spawn(move || {
                    let worker = Box::new(WorkerThread::new_in(ctx));

                    worker.run(barrier);
                })
                .expect("Failed to spawn worker thread");
        }

        {
            let ctx = this.clone();
            let barrier = barrier.clone();

            std::thread::Builder::new()
                .name("heartbeat-thread".to_string())
                .spawn(move || {
                    HeartbeatThread::new(ctx, num_threads).run(barrier);
                })
                .expect("Failed to spawn heartbeat thread");
        }

        barrier.wait();

        this
    }

    pub fn set_should_exit(&self) {
        self.should_exit.store(true, Ordering::Relaxed);
        self.queue.as_sender().broadcast_with(|| Message::Exit);
    }

    pub fn should_exit(&self) -> bool {
        self.should_exit.load(Ordering::Relaxed)
    }

    pub fn new() -> Arc<Self> {
        Self::new_with_threads(crate::util::available_parallelism())
    }

    pub fn global_context() -> &'static Arc<Self> {
        static GLOBAL_CONTEXT: OnceLock<Arc<Context>> = OnceLock::new();

        GLOBAL_CONTEXT.get_or_init(|| Self::new())
    }

    pub fn inject_job(&self, job: SharedJob) {
        self.queue.as_sender().anycast(Message::Shared(job));
    }

    /// Runs closure in this context, processing the other context's worker's jobs while waiting for the result.
    fn run_in_worker_cross<T, F>(self: &Arc<Self>, worker: &WorkerThread, f: F) -> T
    where
        F: FnOnce(&WorkerThread) -> T + Send,
        T: Send,
    {
        // current thread is not in the same context, create a job and inject it into the other thread's context, then wait while working on our jobs.

        // SAFETY: we are waiting on this latch in this thread.
        let _pinned = StackJob::new(move |worker: &WorkerThread| f(worker));
        let job = unsafe { Pin::new_unchecked(&_pinned) };

        let job = Job::from_stackjob(&job);
        unsafe {
            self.inject_job(job.share(Some(worker.receiver.get_token().as_parker())));
        }

        let t = worker.wait_until_recv(job.take_receiver().expect("Job should have a receiver"));

        // touch the job to ensure it is dropped after we are done with it.
        drop(_pinned);

        crate::util::unwrap_or_panic(t)
    }

    /// Run closure in this context, sleeping until the job is done.
    pub fn run_in_worker_cold<T, F>(self: &Arc<Self>, f: F) -> T
    where
        F: FnOnce(&WorkerThread) -> T + Send,
        T: Send,
    {
        // current thread isn't a worker thread, create job and inject into context
        let parker = Parker::new();

        struct CrossJob<F> {
            f: UnsafeCell<ManuallyDrop<F>>,
            _pin: PhantomPinned,
        }

        impl<F> CrossJob<F> {
            fn new(f: F) -> Self {
                Self {
                    f: UnsafeCell::new(ManuallyDrop::new(f)),
                    _pin: PhantomPinned,
                }
            }

            fn into_job<T>(self: &Self) -> Job<T>
            where
                F: FnOnce(&WorkerThread) -> T + Send,
                T: Send,
            {
                Job::from_harness(Self::harness, NonNull::from(&*self).cast())
            }

            unsafe fn unwrap(&self) -> F {
                unsafe { ManuallyDrop::take(&mut *self.f.get()) }
            }

            #[align(8)]
            unsafe fn harness<T>(worker: &WorkerThread, this: NonNull<()>, sender: Option<Sender>)
            where
                F: FnOnce(&WorkerThread) -> T + Send,
                T: Send,
            {
                let this: &CrossJob<F> = unsafe { this.cast().as_ref() };
                let f = unsafe { this.unwrap() };
                let sender: Option<Sender<T>> = unsafe { mem::transmute(sender) };

                let result = catch_unwind(AssertUnwindSafe(|| f(worker)));

                let sender = sender.unwrap();
                unsafe {
                    sender.send_as_ref(result);
                    worker
                        .context
                        .queue
                        .as_sender()
                        .unicast(Message::WakeUp, ReceiverToken::from_parker(sender.parker()));
                }
            }
        }

        let pinned = CrossJob::new(move |worker: &WorkerThread| f(worker));
        let job2 = pinned.into_job();

        self.inject_job(job2.share(Some(&parker)));

        let recv = job2.take_receiver().unwrap();

        let out = crate::util::unwrap_or_panic(recv.recv());

        // touch the job to ensure it is dropped after we are done with it.
        drop(pinned);
        drop(parker);

        out
    }

    /// Run closure in this context.
    #[cfg_attr(feature = "tracing", tracing::instrument(level = "trace", skip_all))]
    pub fn run_in_worker<T, F>(self: &Arc<Self>, f: F) -> T
    where
        T: Send,
        F: FnOnce(&WorkerThread) -> T + Send,
    {
        let _guard = DropGuard::new(|| {
            #[cfg(feature = "tracing")]
            tracing::trace!("run_in_worker: finished");
        });
        match WorkerThread::current_ref() {
            Some(worker) => {
                // check if worker is in the same context
                if Arc::ptr_eq(&worker.context, self) {
                    #[cfg(feature = "tracing")]
                    tracing::trace!("run_in_worker: current thread");
                    f(worker)
                } else {
                    // current thread is a worker for a different context
                    #[cfg(feature = "tracing")]
                    tracing::trace!("run_in_worker: cross-context");
                    self.run_in_worker_cross(worker, f)
                }
            }
            None => {
                // current thread is not a worker for any context
                #[cfg(feature = "tracing")]
                tracing::trace!("run_in_worker: inject into context");
                self.run_in_worker_cold(f)
            }
        }
    }
}

impl Context {
    pub fn spawn<F>(self: &Arc<Self>, f: F)
    where
        F: FnOnce() + Send + 'static,
    {
        let job = Job::from_heapjob(HeapJob::new(|_: &WorkerThread| f()));
        #[cfg(feature = "tracing")]
        tracing::trace!("Context::spawn: spawning job: {:?}", job);
        self.inject_job(job.share(None));
    }

    pub fn spawn_future<T, F>(self: &Arc<Self>, future: F) -> async_task::Task<T>
    where
        F: Future<Output = T> + Send + 'static,
        T: Send + 'static,
    {
        let schedule = move |runnable: Runnable| {
            #[align(8)]
            unsafe fn harness<T>(_: &WorkerThread, this: NonNull<()>, _: Option<Sender>) {
                unsafe {
                    let runnable = Runnable::<()>::from_raw(this);
                    runnable.run();
                }
            }

            let job = Job::<T>::from_harness(harness::<T>, runnable.into_raw());

            self.inject_job(job.share(None));
        };

        let (runnable, task) = unsafe { async_task::spawn_unchecked(future, schedule) };

        runnable.schedule();

        task
    }

    #[allow(dead_code)]
    fn spawn_async<T, Fut, Fn>(self: &Arc<Self>, f: Fn) -> async_task::Task<T>
    where
        Fn: FnOnce() -> Fut + Send + 'static,
        Fut: Future<Output = T> + Send + 'static,
        T: Send + 'static,
    {
        let future = async move { f().await };

        self.spawn_future(future)
    }
}

pub fn run_in_worker<T, F>(f: F) -> T
where
    T: Send,
    F: FnOnce(&WorkerThread) -> T + Send,
{
    Context::global_context().run_in_worker(f)
}

#[cfg(test)]
mod tests {
    use std::sync::atomic::AtomicU8;

    use super::*;

    #[test]
    #[cfg_attr(all(not(miri), feature = "tracing"), tracing_test::traced_test)]
    fn run_in_worker() {
        let ctx = Context::global_context().clone();
        let result = ctx.run_in_worker(|_| 42);
        assert_eq!(result, 42);
    }

    #[test]
    #[cfg_attr(all(not(miri), feature = "tracing"), tracing_test::traced_test)]
    fn context_spawn_future() {
        let ctx = Context::global_context().clone();
        let task = ctx.spawn_future(async { 42 });

        // Wait for the task to complete
        let result = futures::executor::block_on(task);
        assert_eq!(result, 42);
    }

    #[test]
    #[cfg_attr(all(not(miri), feature = "tracing"), tracing_test::traced_test)]
    fn context_spawn_async() {
        let ctx = Context::global_context().clone();
        let task = ctx.spawn_async(|| async { 42 });

        // Wait for the task to complete
        let result = futures::executor::block_on(task);
        assert_eq!(result, 42);
    }

    #[test]
    #[cfg_attr(all(not(miri), feature = "tracing"), tracing_test::traced_test)]
    fn context_spawn() {
        let ctx = Context::global_context().clone();
        let counter = Arc::new(AtomicU8::new(0));
        let barrier = Arc::new(std::sync::Barrier::new(2));

        ctx.spawn({
            let counter = counter.clone();
            let barrier = barrier.clone();
            move || {
                counter.fetch_add(1, Ordering::SeqCst);
                barrier.wait();
            }
        });

        barrier.wait();
        assert_eq!(counter.load(Ordering::SeqCst), 1);
    }

    #[test]
    #[cfg_attr(all(not(miri), feature = "tracing"), tracing_test::traced_test)]
    fn inject_job_and_wake_worker() {
        let ctx = Context::new_with_threads(1);
        let counter = Arc::new(AtomicU8::new(0));

        let parker = Parker::new();

        let job = StackJob::new({
            let counter = counter.clone();
            move |_: &WorkerThread| {
                #[cfg(feature = "tracing")]
                tracing::info!("Job running");
                counter.fetch_add(1, Ordering::SeqCst);

                42
            }
        });

        let job = Job::from_stackjob(&job);

        // wait for the worker to sleep
        std::thread::sleep(std::time::Duration::from_millis(100));

        ctx.heartbeats
            .inner()
            .iter_mut()
            .next()
            .map(|(_, heartbeat)| {
                assert!(heartbeat.is_waiting());
            });

        ctx.inject_job(job.share(Some(&parker)));

        // Wait for the job to be executed
        let recv = job.take_receiver().expect("Job should have a receiver");
        let Some(result) = recv.poll() else {
            panic!("Expected a finished message");
        };

        let result = crate::util::unwrap_or_panic::<i32>(result);
        assert_eq!(result, 42);
        assert_eq!(counter.load(Ordering::SeqCst), 1);
    }
}
